//=============================================================================
// MiniTree.g - Miniature AST Tree Example
//
// Author:  Bary W Pollack
// Date:    Mar. 06, 2004 - 17:00
//
//=============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
}


//-----------------------------------------------------------------------------
// Define a Parser, call it MiniTreeParser
//-----------------------------------------------------------------------------

class MiniTreeParser extends Parser;

options {
    defaultErrorHandler = false;	// Don't generate parser error handlers
    buildAST = true;			// Do construct an AST
}


// Define some methods and variables to use in the generated parser
{
    // Define a main
    public static void main(String[] args) {
        // Use a try/catch block for parser exceptions
        try {
            // if we have at least one command-line argument
            if (args.length > 0 ) {
                System.out.println();
                System.out.println("MiniTree Parsing...");
                System.out.println();

                // For each directory/file specified on the command line
                for (int i = 0; i < args.length; i++)
                    doFile(new File(args[i]));    // parse it
            } else
                System.out.println("Usage: java MiniTreeRecogizer [ file | <directory name> ]...");

        } catch (Exception e) {
            System.out.println("Exception: " + e);
            e.printStackTrace(System.out);     // so we can get a stack trace
        }
    }


    // This method decides what action to take based on the
    // type of file we are looking at -- directory or source file
    public static void doFile(File f) throws Exception {
        // If this is a directory, walk each file/dir in that directory
        if (f.isDirectory()) {
            String files[] = f.list();
            for (int i = 0; i < files.length; i++)
                doFile(new File(f, files[i]));
        // otherwise, if this is a source file, parse it
        } else if ((f.getName().length() >= 5) &&
                   (f.getName().substring(f.getName().length() - 4).equals(".txt") ||
            f.getName().substring(f.getName().length() - 3).equals(".c-"))
          ) {
            //System.out.println("  File: " + f.getAbsolutePath());  //FOO
            System.out.println("  File: " + f.getName());
            parseFile(new FileInputStream(f));
        }
    }

    // Here's where we do the real work...
    public static void parseFile(InputStream s) throws Exception {
        System.out.println();
        try {
            // Create a scanNumer that reads from the input stream passed to us
            MiniTreeLexer lexer = new MiniTreeLexer(s);

            // Create a parser that reads from the scanNumer
            MiniTreeParser parser = new MiniTreeParser(lexer);

            // Start parsing at the program rule
            parser.program();
            System.out.println();
            System.out.println("* Successful Parse *");
	    System.out.println();

	    // Get the tree out of the parser
	    AST resultTree = parser.getAST();

	    // Make an instance of the tree parser
	    SeriesTreeParser treeParser = new SeriesTreeParser();

	    // Begin tree parser at root
	    treeParser.series(resultTree);
	    System.out.println();
	    System.out.println("* Successful Tree Walk *");
	    System.out.println();

	    // Begin tree parser at root
	    treeParser.passTwo(resultTree);
	    System.out.println();
	    System.out.println("* Successful Tree Walk *");
	    // System.out.println();

        } catch (Exception e) {
            System.out.println("Parser exception: " + e);
            System.out.println();
            e.printStackTrace();	// so we can get a stack trace

        } finally {
            System.out.println();
        }
    }

    static void sop(String s) {
        if (true)			//FOO
        System.out.println(s); 
    }

    static int nNum;			// element counter

}


//==================================================================
// MiniTree Syntax
//==================================================================

/*  
 *  Match an element (INT or ID) with possibly a
 *  bunch of ", element" pairs to follow matching
 *  input that looks like 32,a,size,28923,i
 *
 *  A suffixed ! tells the tree generator NOT to create a node
 *  for the matched item
 */

program
{ 
  nNum = 0;
}
    :  element ( COMMA! element { ; } )* DOT!	// No nodes for COMMA or DOT
       { sop(""); sop("There were " + nNum + " elements"); }
    ;

element
{ ++nNum;  }
    :  i:INT   { sop(" " + nNum + " " +  i.getText()); }
    |  id:ID   { sop(" " + nNum + " " + id.getText()); }
    ;


//----------------------------------------------------------------------------
// The MiniTree Scanner
//----------------------------------------------------------------------------

class MiniTreeLexer extends Lexer;

INT   :  ('0'..'9')+ ;

ID    :  ('a'..'z')+ ;

COMMA :  ',' ;

DOT   :  '.' ;


//----------------------------------------------------------------------------
// The MiniTree TreeParser
//----------------------------------------------------------------------------

/*
More complicated translations typically require multiple passes over the input 
and so programmers typically build an intermediate form tree called an abstract 
syntax tree (AST), a structured representation of the text input.  You may walk 
trees by hand or specify an ANTLR tree grammar that describes the tree structure.  
Actions embedded within the tree grammar are executed when the tree parser has 
reached the associated position in the input tree. 

How can we build up a simple tree (albeit a flat one, a list) of the identifier 
and integer input?  Easy.  Tell ANTLR to build trees and it will create a flat 
tree, with one tree node per input token; in other words, ANTLR will build a 
linked list of your input tokens.  To make things a bit more interesting, we 
suffix the COMMA token reference with a '!' to indicate it is not to be 
included in the output tree.

Note that the tree grammar is simpler than the parser grammar.  Normally, you 
will construct trees that are simpler to walk than the text input with all the 
whitespace and other syntactic sugar used to makes things easier for a human 
to read.

You can add actions to a tree grammar just as easily as you can to a parser 
grammar.  To print a list of the integers and identifiers just like in the 
parser grammar, add a few actions (the labels refer to the associated matched 
AST node in tree grammars).
*/


class SeriesTreeParser extends TreeParser;

series
    : (  i:INT   { MiniTreeParser.sop(" " +  i.getText()); }
      |  id:ID   { MiniTreeParser.sop(" " + id.getText()); }
      )+ 
    ;

passTwo
{
    int nSum = 0;
}
    :  (  i:INT  { nSum += Integer.parseInt(i.getText()); }
       |  ID
       )+
       { MiniTreeParser.sop("The sum is " + nSum); }
    ;


// EOF
